package com.oguerreiro.resilient.security.saml2;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.env.Environment;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.saml2.provider.service.authentication.Saml2AuthenticatedPrincipal;
import org.springframework.security.saml2.provider.service.authentication.Saml2Authentication;
import org.springframework.security.web.authentication.AuthenticationFailureHandler;
import org.springframework.security.web.authentication.AuthenticationSuccessHandler;
import org.springframework.stereotype.Component;

import com.oguerreiro.resilient.domain.Organization;
import com.oguerreiro.resilient.repository.OrganizationRepository;
import com.oguerreiro.resilient.repository.security.SecurityGroupRepository;
import com.oguerreiro.resilient.security.custom.ResilientUserDetails;
import com.oguerreiro.resilient.security.resillient.SecurityGroup;

import jakarta.servlet.ServletException;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

@Component
public class Saml2AuthenticationHandler implements AuthenticationSuccessHandler, AuthenticationFailureHandler {

  private static final String ATTR_SECURITYGROUP = "security_group";
  private static final String ATTR_EMAIL = "email";
  private static final String ATTR_ORGANIZATION = "organization_code";
  private static final String ATTR_ROLES = "roles";

  @Autowired
  private Environment environment;

  @Autowired
  private OrganizationRepository organizationRepository;

  @Autowired
  private SecurityGroupRepository securityGroupRepository;

  @Autowired
  private ResilientSaml2Properties resilientSaml2Properties;

  /**
   * AuthenticationSuccessHandler implementation
   */
  @Override
  public void onAuthenticationSuccess(HttpServletRequest request, HttpServletResponse response,
      Authentication authentication) throws IOException, ServletException {
    String samlResponse = request.getParameter("SAMLResponse");
    String samlXMLResponse = Saml2ResponseLoggingFilter.decodeSamlResponse(samlResponse);

    // Read needed information about user security environment
    Saml2AuthenticatedPrincipal principal = (Saml2AuthenticatedPrincipal) authentication.getPrincipal();
    String username = principal.getName();
    String email = getPrincipalAttribute(principal, ATTR_EMAIL);
    String securityGroupCode = getPrincipalAttribute(principal, ATTR_SECURITYGROUP);
    String organizationCode = getPrincipalAttribute(principal, ATTR_ORGANIZATION);
    List<String> roles = principal.getAttribute(ATTR_ROLES);
    Collection<? extends GrantedAuthority> authorities;

    if (roles == null || roles.isEmpty()) {
      authorities = authentication.getAuthorities();
    } else {
      authorities = roles.stream().map(SimpleGrantedAuthority::new).collect(Collectors.toList());
    }

    Organization userOrganization = null;
    SecurityGroup securityGroup = null;

    // Organization exists?
    if (organizationCode == null || organizationCode.isBlank()) {
      // TODO: Make this configurable
      Long defaultOrg = 1L;
      userOrganization = organizationRepository.findById(defaultOrg).orElse(null);
    } else {
      // Find or NULL
      userOrganization = organizationRepository.findOneByCode(organizationCode).orElse(null);
    }
    if (userOrganization == null) {
      this.invalidateLogin(request, response);
      return;
    }

    // SecurityGroup exists?
    String defaultSecurityGroupCode = "default";  // TODO: Make this configurable
    if (securityGroupCode == null || securityGroupCode.isBlank()) {
      securityGroupCode = defaultSecurityGroupCode;
    }

    // Find or fail
    securityGroup = securityGroupRepository.findOneByCodeWithEagerRelationships(securityGroupCode).orElse(null);

    if (securityGroup == null) {
      this.invalidateLogin(request, response);
      return;
    }

    // Create a ResilientUserDetails and replace Principal
    ResilientUserDetails userdetails = new ResilientUserDetails(username, "MOCK-PWD", authorities, securityGroup,
        userOrganization, "pt-PT");

    Saml2Authentication newAuthentication = new Saml2Authentication(userdetails, samlXMLResponse,
        authentication.getAuthorities());
    SecurityContextHolder.getContext().setAuthentication(newAuthentication);

    // This is a sugar-code when in development environment. 
    if (isDevProfileActive()) {
      // If this is a mock-idp, it can provide the parameter 'SAMLDevEnvironmentUrl' 
      //  that gives the base URL to use. This is because in DEV mode usually the 
      //  Angular side runs in localhost:42000 but server-side is in localhost:8080.
      //  Without this, SAMLv2 authentication would end up in error redirecting the user to 
      //  localhost:8080 (NOT the client-side)
      // In PROD we don't need this, because the app url is the same
      String successUrl = resilientSaml2Properties.getSuccessUrl();

      if (successUrl != null && !successUrl.isBlank()) {
        response.sendRedirect(successUrl);
      }
    }
  }

  private void invalidateLogin(HttpServletRequest request, HttpServletResponse response) throws IOException {
    // Clear context and invalidate session
    SecurityContextHolder.clearContext();
    request.getSession(false).invalidate();

    // Send error response or redirect
    response.sendError(HttpServletResponse.SC_UNAUTHORIZED, "Missing or invalid SAML attributes");
  }

  private boolean isDevProfileActive() {
    return Arrays.asList(environment.getActiveProfiles()).contains("dev");
  }

  private String getPrincipalAttribute(Saml2AuthenticatedPrincipal principal, String attribute) {
    List<?> values = principal.getAttribute(attribute);

    if (values == null || values.isEmpty()) {
      return null;
    }

    return principal.getAttribute(attribute).get(0).toString();
  }

  /**
   * AuthenticationFailureHandler implementation
   */
  @Override
  public void onAuthenticationFailure(HttpServletRequest request, HttpServletResponse response,
      AuthenticationException exception) throws IOException, ServletException {

    // This is a sugar-code when in development environment. 
    if (isDevProfileActive()) {
      // If this is a mock-idp, it can provide the parameter 'SAMLDevEnvironmentUrl' 
      //  that gives the base URL to use. This is because in DEV mode usually the 
      //  Angular side runs in localhost:42000 but server-side is in localhost:8080.
      //  Without this, SAMLv2 authentication would end up in error redirecting the user to 
      //  localhost:8080 (NOT the client-side)
      // In PROD we don't need this, because the app url is the same
      String failureUrl = resilientSaml2Properties.getFailureUrl();

      if (failureUrl != null && !failureUrl.isBlank()) {
        response.sendRedirect(failureUrl);
      }
    }
  }
}
